#  _AbstractAlgorithm.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from abc import abstractmethod

from ..core import LoggingObject


class AbstractAlgorithm(LoggingObject):
    """
    Ancestor for algorithms.
    """
    def __init__(self):
        super().__init__()
        self.initialised: bool = False  # Whether the algorithm has been initialised.

    def is_initialised(self):
        """
        Returns whether the algorithm has been trained.

        :return:    True if trained.
        """
        return self.initialised

    def reset(self):
        """
        Resets the scheme.
        """
        super().reset()
        self.initialised = False

    @abstractmethod
    def to_string(self) -> str:
        pass

    def __str__(self):
        return self.to_string()



#  _FFTTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import numpy as np

from wai.test.decorators import RegressionTest, Test
from wai.ma.transformation import FFT, OutputMode
from wai.ma.core.matrix import Matrix

from ..transformation import AbstractTransformationTest


class FFTTest(AbstractTransformationTest):
    @classmethod
    def subject_type(cls):
        return FFT

    @classmethod
    def instantiate_subject(cls, *args, **kwargs) -> FFT:
        fft = FFT()
        fft.output_mode = OutputMode.AMPL_ANGLE_PAIRS
        return fft

    @RegressionTest
    def orthonormal(self, subject: FFT, bolts: Matrix):
        subject.orthonormalisation = True
        return self.standard_regression(subject, bolts)

    @RegressionTest
    def real_imag_pairs(self, subject: FFT, bolts: Matrix):
        subject.output_mode = OutputMode.REAL_AND_IMAG_PAIRS
        return self.standard_regression(subject, bolts)

    @Test
    def inverses(self, subject: FFT, bolts: Matrix):
        # FFT and inverse
        fft: np.ndarray = subject.fft(bolts.data)
        ifft: np.ndarray = subject.ifft(fft)

        # Check if ifft == input.data
        is_equal: bool = np.allclose(ifft, bolts.data)
        self.assertTrue(is_equal, "Full-circle inversion of FFT failed")

        # Internal format and inverse
        internal_format: np.ndarray = subject.internal_format(fft)
        internal_unformat: np.ndarray = subject.internal_unformat(internal_format)

        # Check if internal_unformat == fft
        is_equal: bool = np.allclose(internal_unformat, fft)
        self.assertTrue(is_equal, "Full-circle inversion of internal format failed")

        # Output format and inverse
        output_format: np.ndarray = subject.format_output(internal_format)
        output_unformat: np.ndarray = subject.inverse_format_output(output_format)

        # Check if output_unformat == internal_format
        is_equal: bool = np.allclose(output_unformat, internal_format)
        self.assertTrue(is_equal, "Full-circle inversion of output format failed")



#  _AbstractTransformationTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from abc import ABC
from typing import Tuple

from wai.test.decorators import Test
from wai.ma.core.matrix import Matrix
from wai.ma.transformation import AbstractTransformation

from ..test import AbstractMatrixAlgorithmTest, TestDataset, Tags


class AbstractTransformationTest(AbstractMatrixAlgorithmTest, ABC):
    """
    Abstract tranformation test. Regression for transform and inverse-transform.
    """
    @Test
    def check_inv_transform_eq_input(self, subject: AbstractTransformation, bolts: Matrix):
        transform: Matrix = subject.transform(bolts)
        inverse_transform: Matrix = subject.inverse_transform(transform)

        # Check if input == inverse_transform
        is_equal: bool = bolts.sub(inverse_transform).abs().all(lambda v: v < 1e-7)
        self.assertTrue(is_equal)

    def standard_regression(self, subject: AbstractTransformation, *resources: Matrix):
        bolts, = resources

        subject.configure(bolts)

        transform: Matrix = subject.transform(bolts)
        inverse_transform: Matrix = subject.inverse_transform(transform)

        return {
            Tags.TRANSFORM: transform,
            Tags.INVERSE_TRANSFORM: inverse_transform
        }

    @classmethod
    def get_datasets(cls) -> Tuple[TestDataset]:
        return TestDataset.BOLTS,



#  _GLSWTest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Tuple

from wai.test.decorators import RegressionTest

from wai.ma.algorithm.glsw import GLSW
from wai.ma.core.matrix import Matrix, factory

from ...test import AbstractMatrixAlgorithmTest, Tags, TestDataset


class GLSWTest(AbstractMatrixAlgorithmTest):
    @classmethod
    def subject_type(cls):
        return GLSW

    def add_glsw_regressions(self, subject: GLSW, x: Matrix):
        # Add regressions
        return {
            Tags.TRANSFORM: subject.transform(x),
            Tags.PROJECTION: subject.G
        }

    @RegressionTest
    def alpha_1(self, subject: GLSW, *resources: Matrix):
        subject.alpha = 1
        return self.standard_regression(subject, *resources)

    @RegressionTest
    def alpha_100(self, subject: GLSW, *resources: Matrix):
        subject.alpha = 100
        return self.standard_regression(subject, *resources)

    def standard_regression(self, subject: GLSW, *resources: Matrix):
        # Get inputs: Simulate second instrument as x1 with noise
        x_first_instrument, = resources
        x_second_instrument: Matrix = x_first_instrument.add(factory.randn_like(x_first_instrument, 0, 0.0, 0.1))

        # Init GLSW
        subject.initialize(x_first_instrument, x_second_instrument)

        return self.add_glsw_regressions(subject, x_first_instrument)

    @classmethod
    def get_datasets(cls) -> Tuple[TestDataset]:
        return TestDataset.BOLTS,



#  _Log.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import IO

import numpy as np

from ..core import Serialisable
from ..core.matrix import Matrix
from ._AbstractTransformation import AbstractTransformation


class Log(AbstractTransformation, Serialisable):
    def __init__(self):
        super().__init__()

        self.base: float = np.e
        self.offset: float = 1.0

    def __setattr__(self, key, value):
        # Validate values
        validator_name = 'validate_' + key
        if hasattr(self, validator_name):
            validator = getattr(self, validator_name)
            validator(value)

            # If validation passes, reset
            self.reset()

        super().__setattr__(key, value)

    @staticmethod
    def validate_base(value: float):
        if value == 1.0 or value <= 0.0:
            raise ValueError(f"Logarithmic base must not be 1, 0 or negative (got {value})")

    def configure(self, data: Matrix):
        self.configured = True

    def do_transform(self, data: Matrix) -> Matrix:
        # Copy the matrix data
        data: np.ndarray = data.data.copy()

        # Apply the offset
        data += self.offset

        # Log is undefined for zero and negative values
        undefined: np.ndarray = data <= 0
        if undefined.any():
            raise ValueError("Logarithm is undefined for negative/zero values")

        return Matrix(np.log(data) * self._base_conversion_factor())

    def do_inverse_transform(self, data: Matrix) -> Matrix:
        # Copy the matrix data
        data: np.ndarray = data.data.copy()

        # Undo the logarithm
        data = np.exp(data / self._base_conversion_factor())

        # Undo the offset
        data -= self.offset

        return Matrix(data)

    def _base_conversion_factor(self) -> float:
        """
        Calculates the conversion factor for converting from the natural logarithm
        to logarithms of our base.

        :return:    The conversion factor, 1 / ln(base).
        """
        return 1.0 / np.log(self.base) if self.base != np.e else 1.0

    def serialise_state(self, stream: IO[bytes]):
        # Serialise out our base, base-conversion factor and offset
        stream.write(self.serialise_to_bytes(self.base))
        stream.write(self.serialise_to_bytes(self._base_conversion_factor()))
        stream.write(self.serialise_to_bytes(self.offset))



#  _MultiFilter.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List

from .matrix import Matrix
from ._Filter import Filter


class MultiFilter(Filter):
    """
    Filter which encapsulates a series of sub-filters, and applies
    them in a given order.
    """
    def __init__(self, filters: List[Filter]):
        # The filters in the order of application
        self.filters: List[Filter] = filters
        
    def transform(self, predictors: Matrix) -> Matrix:
        # The result starts as the predictors
        result: Matrix = predictors
        
        # Apply each filter in ordered turn
        for filter in self.filters:
            result = filter.transform(result)
            
        return result



#  _PCATest.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Tuple

from wai.test.decorators import RegressionTest
from wai.ma.algorithm import PCA
from wai.ma.core.matrix import Matrix

from ..test import TestDataset, AbstractMatrixAlgorithmTest, Tags


class PCATest(AbstractMatrixAlgorithmTest):
    @classmethod
    def subject_type(cls):
        return PCA

    @RegressionTest
    def center(self, subject: PCA, bolts: Matrix):
        subject.center = True
        return self.standard_regression(subject, bolts)

    @RegressionTest
    def max_cols_3(self, subject: PCA, bolts: Matrix):
        subject.max_columns = 3
        return self.standard_regression(subject, bolts)

    def standard_regression(self, subject: PCA, *resources: Matrix):
        # Get input
        bolts, = resources

        # Get matrices
        transformed: Matrix = subject.transform(bolts)
        loadings: Matrix = subject.loadings
        scores: Matrix = subject.scores

        # Add regressions
        return {
            Tags.TRANSFORM: transformed,
            Tags.LOADINGS: loadings,
            Tags.SCORES: scores,
        }

    @classmethod
    def get_datasets(cls) -> Tuple[TestDataset]:
        return TestDataset.BOLTS,



#  _VCPLS.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

from ...core import ONE, real
from ...core.matrix import Matrix, factory
from ._PLS1 import PLS1


class VCPLS(PLS1):
    """
    Variance constrained partial least squares

    See also:
    <a href="http://or.nsfc.gov.cn/bitstream/00001903-5/485833/1/1000013952154.pdf">Variance
    constrained partial least squares</a>

    Parameters:
    - lambda: (No description given in paper)
    """
    NU: real = real(1e-7)  # The constant NU.

    def __init__(self):
        super().__init__()
        self.lambda_: real = ONE

    @staticmethod
    def validate_lambda_(value: real) -> bool:
        return True

    def initialize(self, predictors: Optional[Matrix] = None, response: Optional[Matrix] = None) -> Optional[str]:
        if predictors is None and response is None:
            super().initialize()
            self.lambda_ = ONE
        else:
            return super().initialize(predictors, response)

    def calculate_weights(self, x_k: Matrix, y: Matrix) -> Matrix:
        # Paper notation
        e: Matrix = x_k
        f: Matrix = y

        I: Matrix = factory.eye(e.num_columns())
        g_1: Matrix = e.transpose().mul(f).mul(f.transpose()).mul(e).sub(I.mul(self.lambda_))
        g_2: Matrix = e.transpose().mul(e)

        term: Matrix = (g_2.add(I.mul(VCPLS.NU))).inverse().mul(g_1)

        return term.get_dominant_eigenvector()



#  _Filter.py
#  Copyright (C) 2019 University of Waikato, Hamilton, New Zealand
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from abc import ABC, abstractmethod

from .matrix import Matrix


class Filter(ABC):
    """
    Filter API that exposes a method transform(Matrix) which takes
    a matrix and returns a matrix based on this filter's transformation rules.
    """
    @abstractmethod
    def transform(self, predictors: Matrix) -> Matrix:
        """
        Transform a given matrix into another matrix based on the
        filter's implementation.

        :param predictors:  Input matrix.
        :return:            Transformed matrix.
        """
        pass
